#!/usr/bin/perl
## printconf-backend - A Python Printing configuration utility
## Copyright (C) 2000 Red Hat, Inc.
## Copyright (C) 2000 Crutcher Dunnavant <crutcher@redhat.com>,

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


$foomatic_dir = '/usr/share/printconf/foomatic';
$postscript_wrapper = '/usr/share/printconf/util/mf_postscript_wrapper';


undef $/;
umask 0077;
($lpuid, $lpgid) = (getpwnam('lp'))[2,3];

# read in the control structure from stdin
eval (<STDIN>) || exit 1;
$control = $VAR1;

# Could be smarter (and not annoy europe and asia)
$default_page_size = 'Letter';

# Ask mpage what page sizes it supports
open(MPAGE, '/usr/bin/mpage -b? |');
$m_str = <MPAGE>;
close(MPAGE);
$m_str =~ s/^.*\n.*\n.*\n//;
@sizes = split('\n', $m_str);
for $size (@sizes) {
	$size =~ s/^(\S*).*/\1/;
	if ($control->{'page_size'} eq $size) {
		$page_size = $size;
		break;
	}
}
$page_size = $page_size || $default_page_size;
$enscript_cmd = "/usr/bin/mpage -b ifdef(`PAGEsize', PAGEsize, Letter) -1 -o -P- -";

$rerender_ps = ($control->{'flags'}{'rerender_Postscript'} == 1) ? '--rerender-ps' : '';

$termination = ($control->{'flags'}{'send_EOT'}  == 1) ? '\004' : '';

$spool_dir = $control->{'spool_dir'};
$mf_type = $control->{'mf_type'};

if ($mf_type eq 'POSTSCRIPT') {
	# This is a pure postscript printer
	
	#
	# Build the mf.cfg string
	#
	$printer_id = 'Postscript Printer';
	$gs_driver = 'None';

	$text_filter = "pipe/postscript/ $enscript_cmd";
	$ps_filter = "filter $postscript_wrapper $rerender_ps";
	# Yes, this is stupid. But we don't have reliable information anywhere on #dnl
	# if a given printer supports pcl and pjl, so what ya gonna do? #dnl
	$pcl_filter = 'cat';
	$pjl_filter = 'cat';

} elsif ($mf_type eq 'TEXT') {
	# This is a pure text printer
	
	#
	# Build the mf.cfg string
	#
	$printer_id = 'TEXT Printer';
	$gs_driver = 'Dont be silly';

	$text_filter = 'text';
	$ps_filter = 'text';
	$pcl_filter = 'reject';
	$pjl_filter = 'reject';

} elsif ($mf_type eq 'MFOMATIC') {
	# This is a mfomatic printer configuration

	$gs_driver = $control->{'gs_driver'};
	$printer_id = $control->{'printer_id'};

	# Read in the foomatic priter description file
	open (FOO, $pdf_path = "$foomatic_dir/data/$gs_driver-$printer_id.foo") ||
	open (FOO, $pdf_path = "$foomatic_dir/rh-data/$gs_driver-$printer_id.foo") || exit 1;
	eval(<FOO>) or exit 1;
	close FOO;
	$foodata = $VAR1;


	# Read in the charset information, if it exists
	if (open (BAR, "$foomatic_dir/charset_overrides/$printer_id.bar")) {
		eval(<BAR>) and $chardata = $VAR1;
		close BAR;
	}

	#
	# Override the foomatic file
	#
	while (($name, $option) = each %{$control->{'foomatic_defaults'}}) {
		$foo_option = $foodata->{'args_byname'}{$name};
		next if not $foo_option;
	
		$type = $option->{'type'};
		next if not $type eq $foo_option->{'type'};

		$def = $option->{'default'};
		if (($type ne 'enum') or exists $foo_option->{'vals_byname'}{$def}) {
			$foo_option->{'default'} = $def;
		}
	}
	$local_foo_file = "$spool_dir/$gs_driver-$printer_id.foo";
	unlink $local_foo_file;
	open (LFOO, '>', $local_foo_file) or exit 1;
	use Data::Dumper;
	$Data::Dumper::Purity=1;
	$Data::Dumper::Indent=1;
	print LFOO Dumper($foodata);
	close LFOO;
	chown $lpuid, $lpgid, $local_foo_file;

	#
	# Build the mf.cfg string
	#
	$make = $foodata->{'make'};
	$model = $foodata->{'model'};
	$color = ($foodata->{'color'} ? 'true' : 'false');

	$text_filter = $foodata->{'ascii'} ? ($chardata->{'cmd'} || 'text "" \r\n\14 ') : "pipe/postscript/ $enscript_cmd";
	$ps_filter = "filter $postscript_wrapper $rerender_ps --mfomatic -d $gs_driver-$printer_id.foo";
	# Yes, this is stupid. But we don't have reliable information anywhere on #dnl
	# if a given printer supports pcl and pjl, so what ya gonna do? #dnl
	$pcl_filter = 'cat';
	$pjl_filter = 'cat';
} else {
	# I don't understand this format.
	exit 1;
}

$mf_cfg_str = "# foomatic/magicfilter configuration
# Make:		$make
# Model:	$model
# Printer Id:	$printer_id
# Driver:	$gs_driver
# TERMINATION=$termination

define(MAKE, `$make')dnl
define(MODEL, `$model')dnl
define(COLOR, `$color')dnl
define(PAGEsize, `$page_size')dnl
define(TEXTfilter, $text_filter)dnl
define(PSfilter, `$ps_filter')dnl
define(PCLfilter, `$pcl_filter')dnl
define(PJLfilter, `$pjl_filter')dnl
define(DEFAULTfilter, `cat/text/')dnl
";

#
# Write out the mf.cfg file
#
$mf_cfg = $spool_dir . '/mf.cfg';
unlink $mf_cfg;
open (MF_CFG, '>', $mf_cfg) || exit 1;
print MF_CFG $mf_cfg_str;
close MF_CFG;
chown $lpuid, $lpgid, $mf_cfg;


